namespace MyWebLog.Data.SQLite

open System
open Microsoft.Data.Sqlite
open MyWebLog
open MyWebLog.Data

/// SQLite myWebLog user data implementation        
type SQLiteWebLogUserData (conn : SqliteConnection) =
    
    // SUPPORT FUNCTIONS

    /// Add parameters for web log user INSERT or UPDATE statements
    let addWebLogUserParameters (cmd : SqliteCommand) (user : WebLogUser) =
        [ cmd.Parameters.AddWithValue ("@id", WebLogUserId.toString user.id)
          cmd.Parameters.AddWithValue ("@webLogId", WebLogId.toString user.webLogId)
          cmd.Parameters.AddWithValue ("@userName", user.userName)
          cmd.Parameters.AddWithValue ("@firstName", user.firstName)
          cmd.Parameters.AddWithValue ("@lastName", user.lastName)
          cmd.Parameters.AddWithValue ("@preferredName", user.preferredName)
          cmd.Parameters.AddWithValue ("@passwordHash", user.passwordHash)
          cmd.Parameters.AddWithValue ("@salt", user.salt)
          cmd.Parameters.AddWithValue ("@url", maybe user.url)
          cmd.Parameters.AddWithValue ("@accessLevel", AccessLevel.toString user.accessLevel)
          cmd.Parameters.AddWithValue ("@createdOn", user.createdOn)
          cmd.Parameters.AddWithValue ("@lastSeenOn", maybe user.lastSeenOn)
        ] |> ignore
    
    // IMPLEMENTATION FUNCTIONS
    
    /// Add a user
    let add user = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- """
            INSERT INTO web_log_user (
                id, web_log_id, user_name, first_name, last_name, preferred_name, password_hash, salt, url,
                access_level, created_on, last_seen_on
            ) VALUES (
                @id, @webLogId, @userName, @firstName, @lastName, @preferredName, @passwordHash, @salt, @url,
                @accessLevel, @createdOn, @lastSeenOn
            )"""
        addWebLogUserParameters cmd user
        do! write cmd
    }
    
    /// Find a user by their e-mail address for the given web log
    let findByEmail (email : string) webLogId = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- "SELECT * FROM web_log_user WHERE web_log_id = @webLogId AND user_name = @userName"
        addWebLogId cmd webLogId
        cmd.Parameters.AddWithValue ("@userName", email) |> ignore
        use! rdr = cmd.ExecuteReaderAsync ()
        return if rdr.Read () then Some (Map.toWebLogUser rdr) else None
    }
    
    /// Find a user by their ID for the given web log
    let findById userId webLogId = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- "SELECT * FROM web_log_user WHERE id = @id"
        cmd.Parameters.AddWithValue ("@id", WebLogUserId.toString userId) |> ignore
        use! rdr = cmd.ExecuteReaderAsync ()
        return Helpers.verifyWebLog<WebLogUser> webLogId (fun u -> u.webLogId) Map.toWebLogUser rdr 
    }
    
    /// Get all users for the given web log
    let findByWebLog webLogId = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- "SELECT * FROM web_log_user WHERE web_log_id = @webLogId"
        addWebLogId cmd webLogId
        use! rdr = cmd.ExecuteReaderAsync ()
        return toList Map.toWebLogUser rdr
    }
    
    /// Find the names of users by their IDs for the given web log
    let findNames webLogId userIds = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- "SELECT * FROM web_log_user WHERE web_log_id = @webLogId AND id IN ("
        userIds
        |> List.iteri (fun idx userId ->
            if idx > 0 then cmd.CommandText <- $"{cmd.CommandText}, "
            cmd.CommandText <- $"{cmd.CommandText}@id{idx}"
            cmd.Parameters.AddWithValue ($"@id{idx}", WebLogUserId.toString userId) |> ignore)
        cmd.CommandText <- $"{cmd.CommandText})"
        addWebLogId cmd webLogId
        use! rdr = cmd.ExecuteReaderAsync ()
        return
            toList Map.toWebLogUser rdr
            |> List.map (fun u -> { name = WebLogUserId.toString u.id; value = WebLogUser.displayName u })
    }
    
    /// Restore users from a backup
    let restore users = backgroundTask {
        for user in users do
            do! add user
    }
    
    /// Set a user's last seen date/time to now
    let setLastSeen userId webLogId = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- """
            UPDATE web_log_user
               SET last_seen_on = @lastSeenOn
             WHERE id         = @id
               AND web_log_id = @webLogId"""
        addWebLogId cmd webLogId
        [ cmd.Parameters.AddWithValue ("@id", WebLogUserId.toString userId)
          cmd.Parameters.AddWithValue ("@lastSeenOn", DateTime.UtcNow)
        ] |> ignore
        let! _ = cmd.ExecuteNonQueryAsync ()
        ()
    }
    
    /// Update a user
    let update user = backgroundTask {
        use cmd = conn.CreateCommand ()
        cmd.CommandText <- """
            UPDATE web_log_user
               SET user_name      = @userName,
                   first_name     = @firstName,
                   last_name      = @lastName,
                   preferred_name = @preferredName,
                   password_hash  = @passwordHash,
                   salt           = @salt,
                   url            = @url,
                   access_level   = @accessLevel,
                   created_on     = @createdOn,
                   last_seen_on   = @lastSeenOn
             WHERE id         = @id
               AND web_log_id = @webLogId"""
        addWebLogUserParameters cmd user
        do! write cmd
    }
    
    interface IWebLogUserData with
        member _.Add user = add user
        member _.FindByEmail email webLogId = findByEmail email webLogId
        member _.FindById userId webLogId = findById userId webLogId
        member _.FindByWebLog webLogId = findByWebLog webLogId
        member _.FindNames webLogId userIds = findNames webLogId userIds
        member _.Restore users = restore users
        member _.SetLastSeen userId webLogId = setLastSeen userId webLogId
        member _.Update user = update user
